part of '../flutterx_live_data.dart';

abstract class LiveCollection<C> extends LiveData<C> {
  LiveCollection(C source) {
    super.setData(source);
  }

  void setTo(C value);

  @override
  @Deprecated('Use setTo')
  void setData(C value) => setTo(value);
}

class LiveList<E> extends LiveCollection<List<E>> implements List<E> {
  LiveList({List<E>? source}) : super(source ?? <E>[]);

  @override
  bool shouldUpdateValue(List<E> oldValue, List<E> newValue) => !listEquals(oldValue, newValue);

  @override
  void setTo(Iterable<E> value) {
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  bool any(bool Function(E) test) => _data.any(test);

  @override
  List<T> cast<T>() => _data.cast<T>();

  @override
  bool contains(Object? element) => _data.contains(element);

  @override
  E elementAt(int index) => _data.elementAt(index);

  @override
  bool every(bool Function(E) test) => _data.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(E) f) => _data.expand(f);

  @override
  E get first => _data.first;

  @override
  E firstWhere(bool Function(E) test, {E Function()? orElse}) => _data.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, E element) combine) => _data.fold(initialValue, combine);

  @override
  Iterable<E> followedBy(Iterable<E> other) => _data.followedBy(other);

  @override
  void forEach(void Function(E) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterator<E> get iterator => _data.iterator;

  @override
  String join([String separator = '']) => _data.join(separator);

  @override
  E get last => _data.last;

  @override
  E lastWhere(bool Function(E) test, {E Function()? orElse}) => _data.lastWhere(test, orElse: orElse);

  @override
  int get length => _data.length;

  @override
  Iterable<T> map<T>(T Function(E) f) => _data.map(f);

  @override
  E reduce(E Function(E value, E element) combine) => _data.reduce(combine);

  @override
  E get single => _data.single;

  @override
  E singleWhere(bool Function(E) test, {E Function()? orElse}) => _data.singleWhere(test, orElse: orElse);

  @override
  Iterable<E> skip(int n) => _data.skip(n);

  @override
  Iterable<E> skipWhile(bool Function(E) test) => _data.skipWhile(test);

  @override
  Iterable<E> take(int n) => _data.take(n);

  @override
  Iterable<E> takeWhile(bool Function(E) test) => _data.takeWhile(test);

  @override
  List<E> toList({bool growable = true}) => _data.toList(growable: growable);

  @override
  Set<E> toSet() => _data.toSet();

  @override
  Iterable<E> where(bool Function(E) test) => _data.where(test);

  @override
  Iterable<T> whereType<T>() => _data.whereType<T>();

  @override
  String toString() => _data.toString();

  @override
  E operator [](int index) => _data[index];

  @override
  void operator []=(int index, E value) {
    if (identical(_data[index], value)) return;
    _data[index] = value;
    invalidate();
  }

  @override
  List<E> operator +(List<E> other) => _data + other;

  @override
  void add(E value) {
    _data.add(value);
    invalidate();
  }

  @override
  void addAll(Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    _data.addAll(iterable);
    invalidate();
  }

  @override
  Map<int, E> asMap() => _data.asMap();

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  void fillRange(int start, int end, [E? fillValue]) {
    if (start == end) return;
    _data.fillRange(start, end, fillValue);
    invalidate();
  }

  @override
  set first(E value) {
    if (identical(_data.first, value)) return;
    _data.first = value;
    invalidate();
  }

  @override
  Iterable<E> getRange(int start, int end) => _data.getRange(start, end);

  @override
  int indexOf(E element, [int start = 0]) => _data.indexOf(element, start);

  @override
  int indexWhere(bool Function(E) test, [int start = 0]) => _data.indexWhere(test, start);

  @override
  void insert(int index, E element) {
    _data.insert(index, element);
    invalidate();
  }

  @override
  void insertAll(int index, Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    _data.insertAll(index, iterable);
    invalidate();
  }

  @override
  set last(E value) {
    if (identical(_data.last, value)) return;
    _data.last = value;
    invalidate();
  }

  @override
  int lastIndexOf(E element, [int? start]) => _data.lastIndexOf(element, start);

  @override
  int lastIndexWhere(bool Function(E) test, [int? start]) => _data.lastIndexWhere(test, start);

  @override
  set length(int newLength) {
    if (_data.length == newLength) return;
    _data.length = newLength;
    invalidate();
  }

  @override
  bool remove(Object? value) {
    final result = _data.remove(value);
    if (result) invalidate();
    return result;
  }

  @override
  E removeAt(int index) {
    final result = _data.removeAt(index);
    invalidate();
    return result;
  }

  @override
  E removeLast() {
    final result = _data.removeLast();
    invalidate();
    return result;
  }

  @override
  void removeRange(int start, int end) {
    if (start == end) return;
    _data.removeRange(start, end);
    invalidate();
  }

  @override
  void removeWhere(bool Function(E) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  void replaceRange(int start, int end, Iterable<E> iterable) {
    if (start == end) return;
    _data.replaceRange(start, end, iterable);
    invalidate();
  }

  @override
  void retainWhere(bool Function(E) test) {
    final length = _data.length;
    _data.retainWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Iterable<E> get reversed => _data.reversed;

  @override
  void setAll(int index, Iterable<E> iterable) {
    if (index == _data.length || iterable.isEmpty) return;
    _data.setAll(index, iterable);
    invalidate();
  }

  @override
  void setRange(int start, int end, Iterable<E> iterable, [int skipCount = 0]) {
    if (start == end || iterable.isEmpty) return;
    _data.setRange(start, end, iterable, skipCount);
    invalidate();
  }

  @override
  void shuffle([Random? random]) {
    _data.shuffle(random);
    invalidate();
  }

  @override
  void sort([int Function(E, E)? compare]) {
    _data.sort(compare);
    invalidate();
  }

  @override
  List<E> sublist(int start, [int? end]) => _data.sublist(start, end);
}

class MediatorLiveList<E> extends LiveList<E> with MediatorMixin<List<E>> {
  MediatorLiveList({List<E>? source}) : super(source: source);
}

class HistoryLiveList<E> extends LiveList<E> with HistoryMixin<List<E>> {
  HistoryLiveList({List<E>? source}) : super(source: source);
}

class StorageLiveList<E> extends LiveList<E> with StorageMixin<List<E>> {
  @override
  final String key;

  @override
  final Map<String, String> _storage;

  @override
  final List<E> _defaultValue;

  @override
  final String Function(List<E> value) _encode;

  StorageLiveList({
    required this.key,
    required Map<String, String> storage,
    List<E>? defaultValue,
    String Function(List<E> value) encode = _jsonEncodeList,
    List<E> Function(String source)? decode,
  })  : _storage = storage,
        _encode = encode,
        _defaultValue = defaultValue == null ? const [] : List.unmodifiable(defaultValue),
        super(source: StorageMixin.resolve(key, storage, defaultValue, decode ?? _jsonDecodeList));

  static String _jsonEncodeList(List value) => jsonEncode(value);

  static List<E> _jsonDecodeList<E>(String source) => List<E>.from(jsonDecode(source));
}

class LiveMap<K, V> extends LiveCollection<Map<K, V>> implements Map<K, V> {
  LiveMap({Map<K, V>? source}) : super(source ?? <K, V>{});

  @override
  bool shouldUpdateValue(Map<K, V> oldValue, Map<K, V> newValue) => !mapEquals(oldValue, newValue);

  @override
  void setTo(Map<K, V> value) {
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  V? operator [](Object? key) => _data[key];

  @override
  void operator []=(K key, V value) {
    if (identical(_data[key], value)) return;
    _data[key] = value;
    invalidate();
  }

  @override
  void addAll(Map<K, V> other) {
    if (other.isEmpty) return;
    _data.addAll(other);
    invalidate();
  }

  @override
  void addEntries(Iterable<MapEntry<K, V>> entries) {
    if (entries.isEmpty) return;
    _data.addEntries(entries);
    invalidate();
  }

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  Map<K2, V2> cast<K2, V2>() => _data.cast<K2, V2>();

  @override
  bool containsKey(Object? key) => _data.containsKey(key);

  @override
  bool containsValue(Object? value) => _data.containsValue(value);

  @override
  Iterable<MapEntry<K, V>> get entries => _data.entries;

  @override
  void forEach(void Function(K, V) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterable<K> get keys => _data.keys;

  @override
  int get length => _data.length;

  @override
  Map<K2, V2> map<K2, V2>(MapEntry<K2, V2> Function(K, V) transform) => _data.map(transform);

  @override
  V putIfAbsent(K key, V Function() ifAbsent) {
    var updated = false;
    final result = _data.putIfAbsent(key, () {
      updated = true;
      return ifAbsent();
    });
    if (updated) invalidate();
    return result;
  }

  @override
  V? remove(Object? key) {
    final length = _data.length;
    final result = _data.remove(key);
    if (length != _data.length) invalidate();
    return result;
  }

  @override
  void removeWhere(bool Function(K, V) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Iterable<V> get values => _data.values;

  @override
  String toString() => _data.toString();

  @override
  V update(K key, V Function(V) update, {V Function()? ifAbsent}) {
    var updated = false;
    final result = _data.update(key, (value) {
      final newValue = update(value);
      updated = !identical(value, newValue);
      return newValue;
    },
        ifAbsent: ifAbsent == null
            ? null
            : () {
                updated = true;
                return ifAbsent();
              });
    if (updated) invalidate();
    return result;
  }

  @override
  void updateAll(V Function(K, V) update) {
    _data.updateAll(update);
    invalidate();
  }
}

class MediatorLiveMap<K, V> extends LiveMap<K, V> with MediatorMixin<Map<K, V>> {
  MediatorLiveMap({Map<K, V>? source}) : super(source: source);
}

class HistoryLiveMap<K, V> extends LiveMap<K, V> with HistoryMixin<Map<K, V>> {
  HistoryLiveMap({Map<K, V>? source}) : super(source: source);
}

class StorageLiveMap<K, V> extends LiveMap<K, V> with StorageMixin<Map<K, V>> {
  @override
  final String key;

  @override
  final Map<String, String> _storage;

  @override
  final Map<K, V> _defaultValue;

  @override
  final String Function(Map<K, V> value) _encode;

  StorageLiveMap({
    required this.key,
    required Map<String, String> storage,
    Map<K, V>? defaultValue,
    String Function(Map<K, V> value) encode = _jsonEncodeMap,
    Map<K, V> Function(String source)? decode,
  })  : _storage = storage,
        _encode = encode,
        _defaultValue = defaultValue == null ? const {} : Map.unmodifiable(defaultValue),
        super(source: StorageMixin.resolve(key, storage, defaultValue, decode ?? _jsonDecodeMap));

  static String _jsonEncodeMap(Map value) {
    assert(value is Map<String, dynamic>,
        'Default encode only supports instance of Map<String, dynamic>, got $value. Use a custom encode function instead');
    return jsonEncode(value);
  }

  static Map<K, V> _jsonDecodeMap<K, V>(String source) => Map<K, V>.from(jsonDecode(source));
}

class LiveSet<E> extends LiveCollection<Set<E>> implements Set<E> {
  LiveSet({Set<E>? source}) : super(source ?? <E>{});

  @override
  bool shouldUpdateValue(Set<E> oldValue, Set<E> newValue) => !setEquals(oldValue, newValue);

  @override
  void setTo(Iterable<E> value) {
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  bool any(bool Function(E) test) => _data.any(test);

  @override
  Set<T> cast<T>() => _data.cast<T>();

  @override
  bool contains(Object? element) => _data.contains(element);

  @override
  E elementAt(int index) => _data.elementAt(index);

  @override
  bool every(bool Function(E) test) => _data.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(E) f) => _data.expand(f);

  @override
  E get first => _data.first;

  @override
  E firstWhere(bool Function(E) test, {E Function()? orElse}) => _data.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, E element) combine) => _data.fold(initialValue, combine);

  @override
  Iterable<E> followedBy(Iterable<E> other) => _data.followedBy(other);

  @override
  void forEach(void Function(E) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterator<E> get iterator => _data.iterator;

  @override
  String join([String separator = '']) => _data.join(separator);

  @override
  E get last => _data.last;

  @override
  E lastWhere(bool Function(E) test, {E Function()? orElse}) => _data.lastWhere(test, orElse: orElse);

  @override
  int get length => _data.length;

  @override
  Iterable<T> map<T>(T Function(E) f) => _data.map(f);

  @override
  E reduce(E Function(E value, E element) combine) => _data.reduce(combine);

  @override
  E get single => _data.single;

  @override
  E singleWhere(bool Function(E) test, {E Function()? orElse}) => _data.singleWhere(test, orElse: orElse);

  @override
  Iterable<E> skip(int n) => _data.skip(n);

  @override
  Iterable<E> skipWhile(bool Function(E) test) => _data.skipWhile(test);

  @override
  Iterable<E> take(int n) => _data.take(n);

  @override
  Iterable<E> takeWhile(bool Function(E) test) => _data.takeWhile(test);

  @override
  List<E> toList({bool growable = true}) => _data.toList(growable: growable);

  @override
  Set<E> toSet() => _data.toSet();

  @override
  Iterable<E> where(bool Function(E) test) => _data.where(test);

  @override
  Iterable<T> whereType<T>() => _data.whereType<T>();

  @override
  String toString() => _data.toString();

  @override
  bool add(E value) {
    final result = _data.add(value);
    if (result) invalidate();
    return result;
  }

  @override
  void addAll(Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    final length = _data.length;
    _data.addAll(iterable);
    if (length != _data.length) invalidate();
  }

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  bool containsAll(Iterable<Object?> other) => _data.containsAll(other);

  @override
  Set<E> difference(Set<Object?> other) => _data.difference(other);

  @override
  Set<E> intersection(Set<Object?> other) => _data.intersection(other);

  @override
  E? lookup(Object? element) => _data.lookup(element);

  @override
  bool remove(Object? value) {
    final result = _data.remove(value);
    if (result) invalidate();
    return result;
  }

  @override
  void removeAll(Iterable<Object?> elements) {
    final length = _data.length;
    _data.removeAll(elements);
    if (length != _data.length) invalidate();
  }

  @override
  void removeWhere(bool Function(E) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  void retainAll(Iterable<Object?> elements) {
    final length = _data.length;
    _data.retainAll(elements);
    if (length != _data.length) invalidate();
  }

  @override
  void retainWhere(bool Function(E) test) {
    final length = _data.length;
    _data.retainWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Set<E> union(Set<E> other) => _data.union(other);
}

class MediatorLiveSet<E> extends LiveSet<E> with MediatorMixin<Set<E>> {
  MediatorLiveSet({Set<E>? source}) : super(source: source);
}

class HistoryLiveSet<E> extends LiveSet<E> with HistoryMixin<Set<E>> {
  HistoryLiveSet({Set<E>? source}) : super(source: source);
}

class StorageLiveSet<E> extends LiveSet<E> with StorageMixin<Set<E>> {
  @override
  final String key;

  @override
  final Map<String, String> _storage;

  @override
  final Set<E> _defaultValue;

  @override
  final String Function(Set<E> value) _encode;

  StorageLiveSet({
    required this.key,
    required Map<String, String> storage,
    Set<E>? defaultValue,
    String Function(Set<E> value) encode = _jsonEncodeSet,
    Set<E> Function(String source)? decode,
  })  : _storage = storage,
        _encode = encode,
        _defaultValue = defaultValue == null ? const {} : Set.unmodifiable(defaultValue),
        super(source: StorageMixin.resolve(key, storage, defaultValue, decode ?? _jsonDecodeSet));

  static String _jsonEncodeSet(Set value) => jsonEncode(value.toList(growable: false));

  static Set<E> _jsonDecodeSet<E>(String source) => Set<E>.from(jsonDecode(source));
}
