part of '../flutterx_live_data.dart';

class MutableLiveData<T> extends LiveData<T> {
  MutableLiveData({T? initialValue}) : assert(initialValue is T, 'initialValue must be of type $T, got $initialValue') {
    setData(initialValue as T);
  }

  MutableLiveData.late();

  set value(T value) => setData(value);
}

class MutableMediatorLiveData<T> extends MutableLiveData<T> with MediatorMixin<T> {
  MutableMediatorLiveData({T? initialValue}) : super(initialValue: initialValue);

  MutableMediatorLiveData.late() : super.late();
}

class HistoryMutableLiveData<T> extends MutableLiveData<T> with HistoryMixin<T> {
  HistoryMutableLiveData({T? initialValue}) : super(initialValue: initialValue);

  HistoryMutableLiveData.late() : super.late();
}

class StorageMutableLiveData<T> extends MutableLiveData<T> with StorageMixin<T> {
  @override
  final String key;

  @override
  final Map<String, String> _storage;

  @override
  final T _defaultValue;

  @override
  final String Function(T value) _encode;

  StorageMutableLiveData({
    required this.key,
    required Map<String, String> storage,
    T? defaultValue,
    String Function(T value) encode = _defaultEncode,
    T Function(String source)? decode,
  })  : assert(defaultValue is T, 'defaultValue must be of type $T, got $defaultValue'),
        _storage = storage,
        _encode = encode,
        _defaultValue = defaultValue as T,
        super(initialValue: StorageMixin.resolve(key, storage, defaultValue, decode ?? _defaultDecode));

  static final c = StorageMutableLiveData<String?>.json(key: '', storage: {}, fromJson: (_) => null, toJson: (n) => {});

  StorageMutableLiveData.json({
    required String key,
    required Map<String, String> storage,
    T? defaultValue,
    dynamic Function(T value)? toJson,
    required T Function(dynamic json) fromJson,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: (value) => jsonEncode(toJson == null ? value : toJson(value)),
            decode: (source) => fromJson(jsonDecode(source)));

  static StorageMutableLiveData<int> int_({
    required String key,
    required Map<String, String> storage,
    int defaultValue = 0,
  }) =>
      StorageMutableLiveData<int>(key: key, storage: storage, defaultValue: defaultValue, decode: int.parse);

  static StorageMutableLiveData<double> double_({
    required String key,
    required Map<String, String> storage,
    double defaultValue = 0,
  }) =>
      StorageMutableLiveData<double>(key: key, storage: storage, defaultValue: defaultValue, decode: double.parse);

  static StorageMutableLiveData<bool> bool_({
    required String key,
    required Map<String, String> storage,
    bool defaultValue = false,
  }) =>
      StorageMutableLiveData<bool>(
          key: key, storage: storage, defaultValue: defaultValue, decode: (source) => 'true' == source);

  static String _defaultEncode(dynamic value) => value.toString();

  static T _defaultDecode<T>(String source) {
    assert(source is T,
        'Default decode only supports ${source.runtimeType}, got $T. Use a custom decode function instead');
    return source as T;
  }
}
