part of '../flutterx_live_data.dart';

mixin HistoryMixin<T> on LiveData<T> {
  final Queue<String> _queue = Queue();
  int _maxStates = 20;
  int _currentIndex = -1;
  bool _restoring = false;

  final LiveData<bool> canUndo = MutableLiveData(initialValue: false);

  final LiveData<bool> canRedo = MutableLiveData(initialValue: false);

  @override
  void setData(T value) {
    super.setData(value);
    _restoring = false;
  }

  @override
  void invalidate() {
    super.invalidate();
    if (!_restoring) _saveState();
  }

  int get maxStates => _maxStates;

  set maxStates(int value) {
    _maxStates = value;
    while (_queue.length >= value) _queue.removeFirst();
  }

  int get historyIndex => _currentIndex;

  int get historyLength => _queue.length;

  void historyUndo() {
    assert(canUndo.value, 'Cannot undo');
    _currentIndex--;
    _restoreState();
    _updateValues();
  }

  void historyRedo() {
    assert(canRedo.value, 'Cannot redo');
    _currentIndex++;
    _restoreState();
    _updateValues();
  }

  void historyClear() {
    _queue.clear();
    _currentIndex = -1;
    _updateValues();
  }

  @protected
  String serialize(T data) => jsonEncode(data);

  @protected
  T deserialize(String data) => jsonDecode(data);

  void _saveState() {
    if (!initialized) return;
    final state = serialize(_data);
    while (_queue.length > _currentIndex + 1) _queue.removeLast();
    if (_queue.isEmpty || state != _queue.last) {
      _queue.add(state);
      _currentIndex == _maxStates - 1 ? _queue.removeFirst() : _currentIndex++;
      _updateValues();
    }
  }

  void _restoreState() {
    _restoring = true;
    setData(deserialize(_queue.elementAt(_currentIndex)));
  }

  void _updateValues() {
    (canUndo as MutableLiveData).value = _currentIndex > 0;
    (canRedo as MutableLiveData).value = _currentIndex < _queue.length - 1;
  }
}
