part of '../flutterx_live_data.dart';

/// Live data that stores and reads it's value in from a storage if the value differs from default value
mixin StorageMixin<T> on LiveData<T> {
  String get key;

  Map<String, String> get _storage;

  T get _defaultValue;

  String Function(T value) get _encode;

  @override
  void invalidate() {
    super.invalidate();
    if (shouldUpdateValue(_defaultValue, _data))
      _storage[key] = _encode(_data);
    else
      _storage.remove(key);
  }

  void reset() => setData(_defaultValue);

  static T resolve<T>(String key, Map<String, String> storage, T defaultValue, T Function(String source) decode) {
    final value = storage[key];
    return value == null ? defaultValue : decode(value);
  }
}
