part of '../flutterx_live_data.dart';

mixin StateObserver<T extends StatefulWidget> on State<T> {
  final Map<int, ObserverWrapper> _disposables = {};
  int _key = 0;

  @override
  void initState() {
    super.initState();
    _doRegister();
  }

  @override
  void didUpdateWidget(T oldWidget) {
    super.didUpdateWidget(oldWidget);
    _doRegister();
  }

  @override
  void dispose() {
    super.dispose();
    for (final disposable in _disposables.values) disposable.dispose();
  }

  void registerObservers(BuildContext context);

  void observeAll(Iterable<LiveData> sources, void Function(List values) onChanged) =>
      _observeAllInternal(sources, onChanged, _observe);

  void _doRegister() => WidgetsBinding.instance!.addPostFrameCallback((_) {
        _key = 0;
        registerObservers(context);
        assert(
            _disposables.length == _key, 'all observables inside registerObservers must be observed unconditionally');
      });

  void _observe<V>(LiveData<V> data, Observer<V> onChanged) {
    final key = _key++;
    final existing = _disposables[key];
    if (existing != null) {
      if (existing.updated(data, onChanged)) return;
      existing.dispose();
    }
    _disposables[key] = data.observeForever(onChanged);
  }
}
